@@ -6,10 +6,10 @@
 // copied, modified, or distributed except according to those terms.
 
 use core::cell::UnsafeCell;
-use core::ops::{Deref, DerefMut};
 use core::fmt;
-use core::mem;
 use core::marker::PhantomData;
+use core::mem;
+use core::ops::{Deref, DerefMut};
 
 #[cfg(feature = "owning_ref")]
 use owning_ref::StableAddress;
@@ -187,8 +187,7 @@ pub unsafe trait RawRwLockUpgradeFair: RawRwLockUpgrade + RawRwLockFair {
 
 /// Additional methods for RwLocks which support upgradable locks and lock
 /// downgrading.
-pub unsafe trait RawRwLockUpgradeDowngrade
-    : RawRwLockUpgrade + RawRwLockDowngrade {
+pub unsafe trait RawRwLockUpgradeDowngrade: RawRwLockUpgrade + RawRwLockDowngrade {
     /// Downgrades an upgradable lock to a shared lock.
     fn downgrade_upgradable(&self);
 
@@ -354,13 +353,13 @@ impl<R: RawRwLock, T: ?Sized> RwLock<R, T> {
     }
 
     /// Forcibly unlocks a read lock.
-    /// 
+    ///
     /// This is useful when combined with `mem::forget` to hold a lock without
     /// the need to maintain a `RwLockReadGuard` object alive, for example when
     /// dealing with FFI.
-    /// 
+    ///
     /// # Safety
-    /// 
+    ///
     /// This method must only be called if the current thread logically owns a
     /// `RwLockReadGuard` but that guard has be discarded using `mem::forget`.
     /// Behavior is undefined if a rwlock is read-unlocked when not read-locked.
@@ -370,13 +369,13 @@ impl<R: RawRwLock, T: ?Sized> RwLock<R, T> {
     }
 
     /// Forcibly unlocks a write lock.
-    /// 
+    ///
     /// This is useful when combined with `mem::forget` to hold a lock without
     /// the need to maintain a `RwLockWriteGuard` object alive, for example when
     /// dealing with FFI.
-    /// 
+    ///
     /// # Safety
-    /// 
+    ///
     /// This method must only be called if the current thread logically owns a
     /// `RwLockWriteGuard` but that guard has be discarded using `mem::forget`.
     /// Behavior is undefined if a rwlock is write-unlocked when not write-locked.
@@ -402,13 +401,13 @@ impl<R: RawRwLock, T: ?Sized> RwLock<R, T> {
 
 impl<R: RawRwLockFair, T: ?Sized> RwLock<R, T> {
     /// Forcibly unlocks a read lock using a fair unlock procotol.
-    /// 
+    ///
     /// This is useful when combined with `mem::forget` to hold a lock without
     /// the need to maintain a `RwLockReadGuard` object alive, for example when
     /// dealing with FFI.
-    /// 
+    ///
     /// # Safety
-    /// 
+    ///
     /// This method must only be called if the current thread logically owns a
     /// `RwLockReadGuard` but that guard has be discarded using `mem::forget`.
     /// Behavior is undefined if a rwlock is read-unlocked when not read-locked.
@@ -418,13 +417,13 @@ impl<R: RawRwLockFair, T: ?Sized> RwLock<R, T> {
     }
 
     /// Forcibly unlocks a write lock using a fair unlock procotol.
-    /// 
+    ///
     /// This is useful when combined with `mem::forget` to hold a lock without
     /// the need to maintain a `RwLockWriteGuard` object alive, for example when
     /// dealing with FFI.
-    /// 
+    ///
     /// # Safety
-    /// 
+    ///
     /// This method must only be called if the current thread logically owns a
     /// `RwLockWriteGuard` but that guard has be discarded using `mem::forget`.
     /// Behavior is undefined if a rwlock is write-unlocked when not write-locked.
@@ -670,9 +669,7 @@ impl<R: RawRwLock, T> From<T> for RwLock<R, T> {
 impl<R: RawRwLock, T: ?Sized + fmt::Debug> fmt::Debug for RwLock<R, T> {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result {
         match self.try_read() {
-            Some(guard) => f.debug_struct("RwLock")
-                .field("data", &&*guard)
-                .finish(),
+            Some(guard) => f.debug_struct("RwLock").field("data", &&*guard).finish(),
             None => f.pad("RwLock { <locked> }"),
         }
     }
@@ -717,7 +714,9 @@ impl<'a, R: RawRwLock + 'a, T: ?Sized + 'a> RwLockReadGuard<'a, R, T> {
         }
     }
 
-    /// Executes the given function with the `RwLock` unlocked.
+    /// Temporarily unlocks the `RwLock` to execute the given function.
+    ///
+    /// The `RwLock` is unlocked a fair unlock protocol.
     ///
     /// This is safe because `&mut` guarantees that there exist no other
     /// references to the data protected by the `RwLock`.
@@ -751,8 +750,9 @@ impl<'a, R: RawRwLockFair + 'a, T: ?Sized + 'a> RwLockReadGuard<'a, R, T> {
         mem::forget(s);
     }
 
-    /// Executes the given function with the `RwLock` unlocked using a fair unlock
-    /// protocol.
+    /// Temporarily unlocks the `RwLock` to execute the given function.
+    ///
+    /// The `RwLock` is unlocked a fair unlock protocol.
     ///
     /// This is safe because `&mut` guarantees that there exist no other
     /// references to the data protected by the `RwLock`.
@@ -792,6 +792,17 @@ impl<'a, R: RawRwLock + 'a, T: ?Sized + 'a> Drop for RwLockReadGuard<'a, R, T> {
     }
 }
 
+impl<'a, R: RawRwLockRecursive + 'a, T: ?Sized + 'a> Clone for RwLockReadGuard<'a, R, T> {
+    #[inline]
+    fn clone(&self) -> Self {
+        self.rwlock.raw.lock_shared_recursive();
+        RwLockReadGuard {
+            rwlock: self.rwlock,
+            marker: PhantomData,
+        }
+    }
+}
+
 #[cfg(feature = "owning_ref")]
 unsafe impl<'a, R: RawRwLock + 'a, T: ?Sized + 'a> StableAddress for RwLockReadGuard<'a, R, T> {}
 
@@ -834,7 +845,7 @@ impl<'a, R: RawRwLock + 'a, T: ?Sized + 'a> RwLockWriteGuard<'a, R, T> {
         }
     }
 
-    /// Executes the given function with the `RwLock` unlocked.
+    /// Temporarily unlocks the `RwLock` to execute the given function.
     ///
     /// This is safe because `&mut` guarantees that there exist no other
     /// references to the data protected by the `RwLock`.
@@ -904,8 +915,9 @@ impl<'a, R: RawRwLockFair + 'a, T: ?Sized + 'a> RwLockWriteGuard<'a, R, T> {
         mem::forget(s);
     }
 
-    /// Executes the given function with the `RwLock` unlocked using a fair unlock
-    /// protocol.
+    /// Temporarily unlocks the `RwLock` to execute the given function.
+    ///
+    /// The `RwLock` is unlocked a fair unlock protocol.
     ///
     /// This is safe because `&mut` guarantees that there exist no other
     /// references to the data protected by the `RwLock`.
@@ -965,8 +977,7 @@ pub struct RwLockUpgradableReadGuard<'a, R: RawRwLockUpgrade + 'a, T: ?Sized + '
 
 unsafe impl<'a, R: RawRwLockUpgrade + 'a, T: ?Sized + Sync + 'a> Sync
     for RwLockUpgradableReadGuard<'a, R, T>
-{
-}
+{}
 
 impl<'a, R: RawRwLockUpgrade + 'a, T: ?Sized + 'a> RwLockUpgradableReadGuard<'a, R, T> {
     /// Returns a reference to the original reader-writer lock object.
@@ -974,7 +985,7 @@ impl<'a, R: RawRwLockUpgrade + 'a, T: ?Sized + 'a> RwLockUpgradableReadGuard<'a,
         s.rwlock
     }
 
-    /// Executes the given function with the `RwLock` unlocked.
+    /// Temporarily unlocks the `RwLock` to execute the given function.
     ///
     /// This is safe because `&mut` guarantees that there exist no other
     /// references to the data protected by the `RwLock`.
@@ -1036,8 +1047,9 @@ impl<'a, R: RawRwLockUpgradeFair + 'a, T: ?Sized + 'a> RwLockUpgradableReadGuard
         mem::forget(s);
     }
 
-    /// Executes the given function with the `RwLock` unlocked using a fair unlock
-    /// protocol.
+    /// Temporarily unlocks the `RwLock` to execute the given function.
+    ///
+    /// The `RwLock` is unlocked a fair unlock protocol.
     ///
     /// This is safe because `&mut` guarantees that there exist no other
     /// references to the data protected by the `RwLock`.
@@ -1144,8 +1156,7 @@ impl<'a, R: RawRwLockUpgrade + 'a, T: ?Sized + 'a> Drop for RwLockUpgradableRead
 #[cfg(feature = "owning_ref")]
 unsafe impl<'a, R: RawRwLockUpgrade + 'a, T: ?Sized + 'a> StableAddress
     for RwLockUpgradableReadGuard<'a, R, T>
-{
-}
+{}
 
 /// An RAII read lock guard returned by `RwLockReadGuard::map`, which can point to a
 /// subfield of the protected data.
@@ -1162,11 +1173,9 @@ pub struct MappedRwLockReadGuard<'a, R: RawRwLock + 'a, T: ?Sized + 'a> {
 }
 
 unsafe impl<'a, R: RawRwLock + 'a, T: ?Sized + Sync + 'a> Sync for MappedRwLockReadGuard<'a, R, T> {}
-unsafe impl<'a, R: RawRwLock + 'a, T: ?Sized + 'a> Send for MappedRwLockReadGuard<'a, R, T>
-where
-    R::GuardMarker: Send,
-{
-}
+unsafe impl<'a, R: RawRwLock + 'a, T: ?Sized + 'a> Send for MappedRwLockReadGuard<'a, R, T> where
+    R::GuardMarker: Send
+{}
 
 impl<'a, R: RawRwLock + 'a, T: ?Sized + 'a> MappedRwLockReadGuard<'a, R, T> {
     /// Make a new `MappedRwLockReadGuard` for a component of the locked data.
@@ -1228,11 +1237,22 @@ impl<'a, R: RawRwLock + 'a, T: ?Sized + 'a> Drop for MappedRwLockReadGuard<'a, R
     }
 }
 
+impl<'a, R: RawRwLockRecursive + 'a, T: ?Sized + 'a> Clone for MappedRwLockReadGuard<'a, R, T> {
+    #[inline]
+    fn clone(&self) -> Self {
+        self.raw.lock_shared_recursive();
+        MappedRwLockReadGuard {
+            raw: self.raw,
+            data: self.data,
+            marker: PhantomData,
+        }
+    }
+}
+
 #[cfg(feature = "owning_ref")]
 unsafe impl<'a, R: RawRwLock + 'a, T: ?Sized + 'a> StableAddress
     for MappedRwLockReadGuard<'a, R, T>
-{
-}
+{}
 
 /// An RAII write lock guard returned by `RwLockWriteGuard::map`, which can point to a
 /// subfield of the protected data.
@@ -1250,13 +1270,10 @@ pub struct MappedRwLockWriteGuard<'a, R: RawRwLock + 'a, T: ?Sized + 'a> {
 
 unsafe impl<'a, R: RawRwLock + 'a, T: ?Sized + Sync + 'a> Sync
     for MappedRwLockWriteGuard<'a, R, T>
-{
-}
-unsafe impl<'a, R: RawRwLock + 'a, T: ?Sized + 'a> Send for MappedRwLockWriteGuard<'a, R, T>
-where
-    R::GuardMarker: Send,
-{
-}
+{}
+unsafe impl<'a, R: RawRwLock + 'a, T: ?Sized + 'a> Send for MappedRwLockWriteGuard<'a, R, T> where
+    R::GuardMarker: Send
+{}
 
 impl<'a, R: RawRwLock + 'a, T: ?Sized + 'a> MappedRwLockWriteGuard<'a, R, T> {
     /// Make a new `MappedRwLockWriteGuard` for a component of the locked data.
@@ -1348,5 +1365,4 @@ impl<'a, R: RawRwLock + 'a, T: ?Sized + 'a> Drop for MappedRwLockWriteGuard<'a,
 #[cfg(feature = "owning_ref")]
 unsafe impl<'a, R: RawRwLock + 'a, T: ?Sized + 'a> StableAddress
     for MappedRwLockWriteGuard<'a, R, T>
-{
-}
+{}
